# rstCA_v9.2.py
# Description: CA simulation platform using raster dataset (0-1 sim)
# Requirements: ArcGIS 9.2 with spatial analysis extensions
# Author: Ying LONG, longying1980@gmail.com, Tsinghua University&BICP
# All right reserved, please do not ditribute it without author's permission
# Updated Date: 17/1/2009

# TODO: polish max method; 

import sys, string, random, os, time, arcgisscripting
gp = arcgisscripting.create()
gp.OverwriteOutput = 1

try:
    gp.Workspace = r"E:\FSA\SIM"
    logFile = r"E:\FSA\SIM\Result.mdb\LOG"
    method = "Normal"
    loops = 1    #MonoLoop
    dtN =   3    #MonoLoop
    N0  =  10    #MonoLoop
    steps = 200
    alfa = 2
    k = 20.0

    lstA = [ 9254,25, -8.700,15.268, 3.575,-0.717, 4.105,1.368, 1.193,-2.396,15]
    lstB = [ 9270,25,-30.696,54.558,10.294, 5.272, 8.765,6.027, 3.672, 5.066,17]
    lstC = [ 9895,29,-63.599,15.106,10.046,31.639,24.348,7.627, 4.078, 6.094, 9]
    lstD = [10679,34,-55.624,20.849, 9.701, 7.807,11.622,8.113,23.000,12.003, 7]
    lstH = [ 5011, 9,-12.263,11.782, 2.490,-1.872, 7.574,0.917, 1.535,-1.179,20]

    lstFM = lstC            # Select the form scenario
    fm_obs = "fm_C"         # Select the form to be compared
    fm_start = "fm_2006"
    fm_extent = "full"

    inThreshold = 0.99      # Normal method
    stepNum = lstFM[1]      # Max method
    
    w0=lstFM[2]  #Constant
    w1=lstFM[3]  #tam
    w2=lstFM[4]  #city
    w3=lstFM[5]  #town
    w4=lstFM[6]  #river
    w5=lstFM[7]  #road
    w6=lstFM[8]  #conf
    w7=lstFM[9]  #agri
    wN0=lstFM[10]#neighbor
    
    r1="l_tam"
    r2="l_city"
    r3="l_town"
    r4="l_river"
    r5="l_road"
    r6="g_conf"
    r7="g_agri"

    gp.CheckOutExtension("Spatial")
    # Delete former log
    cur = gp.UpdateCursor(logFile)
    row = cur.Next()
    while row:
        cur.DeleteRow(row)
        row = cur.Next()
    del row, cur
    print("Former log records deleted")
    
    # Delete former outputs and suit
    rsList = gp.ListRasters("ld*", "GRID")
    rs = rsList.Next()
    while rs:
        gp.delete(rs)
        print (rs + " deleted")
        rs = rsList.Next()
    del rs,rsList
    if gp.exists("suit"):
        gp.delete("suit")
        print ("suit deleted")
    gp.RefreshCatalog(gp.Workspace)

    gp.ZonalStatisticsAsTable_sa(fm_extent, "Value", fm_start, "zonestats.dbf", "DATA")
    # gp.ZonalStatisticsAsTable_sa(inRaster1, "Value", inRaster2, outTable, "DATA")
    
    cur = gp.SearchCursor("zonestats.dbf")
    row = cur.next()
    numStart0 = int(row.SUM)
    del cur,row
    gp.RefreshCatalog(gp.Workspace)
    gp.delete("zonestats.dbf")
    gp.RefreshCatalog(gp.Workspace)

    # Calculate suit
    expr = str(w0)+" + "+str(w1)+" * "+r1+" + "+str(w2)+" * "+r2+" + "+str(w3)+" * "+r3+" + "+str(w4)+" * "+r4
    expr = expr+" + "+str(w5)+" * "+r5+" + "+str(w6)+" * "+r6+" + "+str(w7)+" * "+r7
    gp.SingleOutputMapAlgebra_sa(expr, "suit")
    gp.RefreshCatalog(gp.Workspace)
    print(expr)

    startT = time.time()
    alfa0 = alfa
    ####MonoLoop END
    for j in range(loops):
        if not(loops==1):
            print ("Loop ID is: " + str(j + 0) + "-------------------------------------")
            alfa = alfa0
            wN = N0 + dtN*(j+1)
            print("wN = " + str(wN))
        else:
            wN = wN0
        
        for i in range(steps):
            print ("Iteration starts: " + str(i + 1) + "------------------------")
            # Delete all land* rasters in output directory
            rsList = gp.ListRasters("land*", "GRID")
            rs = rsList.Next()
            while rs:
                gp.delete(rs)
                rs = rsList.Next()
            del rs,rsList
            gp.RefreshCatalog(gp.Workspace)

            if i == 0:
                inFC = fm_start
                numStart = numStart0
            else:
                inFC = "ld" + str(i)
                numStart = numEnd
            print ("numStart = " + str(numStart))

            # rN calculate
            rN_t = "land_nt"
            InNeighborhood = "RECTANGLE 3 3 CELL"
            gp.FocalStatistics_sa(inFC, rN_t, InNeighborhood, "SUM", "DATA")
            gp.RefreshCatalog(gp.Workspace)
            rN = "( " + rN_t + " - " + inFC + " ) / 8"
            
            # p0 calculate
            expr = " suit + " + str(wN) + " * " + rN
            # p1 calculate
            expr = " 1 / ( 1 + exp ( -1 * ( " + expr + " ) ) )"
            expr = "( 1 - " + inFC + " ) * " + expr
            gp.SingleOutputMapAlgebra_sa(expr, "land_p1")
            # cal maximum of p1
            gp.ZonalStatisticsAsTable_sa(fm_extent, "Value", "land_p1", "zonestats.dbf", "DATA")
            cur = gp.SearchCursor("zonestats.dbf")
            row = cur.next()
            pMax = row.MAX
            del cur,row
            gp.RefreshCatalog(gp.Workspace)
            gp.delete("zonestats.dbf")
            gp.RefreshCatalog(gp.Workspace)

            # p2 calculate
            expr = " exp ( " + str(alfa) + " * ( " + " land_p1 " + " / " + str(pMax) + " - 1) "
            # expr = expr + " * ( 1 + ( rand() - 0.5 ) / " + str(k) + " ) )"  # RI item
            expr = expr + " )"
            
            ########difference START
            # threshold control
            if method == "Normal":
                expr = expr + " - " + str(inThreshold) + " "
                gp.SingleOutputMapAlgebra_sa(expr, "land_p2")
                gp.RefreshCatalog(gp.Workspace)

                gp.GreaterThanEqual_sa("land_p2", 0, "land_R")
                gp.RefreshCatalog(gp.Workspace)
            else:   # Max method
                # p2 re-calculate
                expr = "( 1 - " + inFC + " ) * " + expr
                gp.SingleOutputMapAlgebra_sa(expr,"land_p2")
                gp.Times_sa(fm_extent,0,"land_R")
                for k in range(stepNum):
                    # calculate max of P2
                    print("__stepNum is " + str(k+1))
                    gp.CalculateStatistics_management("land_p2")
                    gp.ZonalStatisticsAsTable_sa(fm_extent, "Value", "land_p2", "zonestats.dbf", "DATA")
                    cur = gp.SearchCursor("zonestats.dbf")
                    row = cur.next()
                    pMax = float(row.MAX)
                    del cur,row
                    gp.RefreshCatalog(gp.Workspace)
                    gp.delete("zonestats.dbf")
                    gp.RefreshCatalog(gp.Workspace)
                    print ("  max P2 = " + str(pMax))

                    # gp.EqualTo_sa(pMax, "land_p2", "land_Ri")
                    expr = " Abs ( " + str(pMax) + " - land_p2 ) < 0.00001 "
                    gp.SingleOutputMapAlgebra_sa(expr,"land_Ri")
                    expr = " land_p2 - " + str(pMax) + " * land_Ri "
                    gp.SingleOutputMapAlgebra_sa(expr, "land_pt")
                    gp.delete("land_p2")
                    gp.RefreshCatalog(gp.Workspace)
                    gp.Times_sa("land_pt",1,"land_p2")
                    gp.RefreshCatalog(gp.Workspace)
                    
                    gp.Plus_sa("land_R", "land_Ri", "land_Rt")
                    gp.delete("land_R")
                    gp.RefreshCatalog(gp.Workspace)
                    gp.Times_sa("land_Rt",1,"land_R")
                    gp.RefreshCatalog(gp.Workspace)
            ########difference END
            outFC = "ld" + str(i+1)
            gp.BooleanOr_sa("land_R", inFC, outFC)
            gp.RefreshCatalog(gp.Workspace)

            # calculate Kappa and GOF
            gp.ZonalStatisticsAsTable_sa(fm_obs, "Value", outFC, "zonestats.dbf", "DATA")
            gp.RefreshCatalog(gp.Workspace)
            cur = gp.SearchCursor("zonestats.dbf")
            row = cur.next()
            x11 = int(row.COUNT)-int(row.SUM)
            x12 = int(row.SUM)
            row = cur.next()
            x21 = int(row.COUNT)-int(row.SUM)
            x22 = int(row.SUM)
            del cur,row
            gp.delete("zonestats.dbf")
            gp.RefreshCatalog(gp.Workspace)

            xT = float(x11+x12+x21+x22)
            x1 = x11 + x12
            x2 = x21 + x22
            x_1 = x11 + x21
            x_2 = x12 + x22
            Kappa = float((xT*(x11+x22)-x1*x_1-x2*x_2)/(xT*xT-x1*x_1-x2*x_2))*100.0
            GOF = float((x11+x22)/xT)*100.0
            print("Kappa = " + str(Kappa))
            print("GOF = " + str(GOF))
            
            # Calculate total built-up
            numEnd = x12+x22
            print ("numEnd = " + str(numEnd))
            numCHG = numEnd - numStart
            print ("numCHG = " + str(numCHG))

            # change alfa according to numGap, for Normal method
            if numCHG < 20 and method == "Normal":
                alfa = alfa * 0.9
                print ("alfa = " + str(alfa))
           
            cur = gp.InsertCursor(logFile)
            row = cur.NewRow()
            row.LoopID = j + 1
            row.StepID = i + 1
            row.xN = wN
            row.numStart = numStart
            row.numEnd = numEnd
            row.numCHG = numCHG
            row.alfa = alfa
            row.GOF = GOF
            row.Kappa = Kappa
            cur.InsertRow(row)
            print("New log record inserted")

            numStart = numEnd
            print ("Iteration finished: " + str(i + 1))

            if numEnd > x2: #9254
                print("Breaking.....................")
                break
            
        # Delete no use rasters in the directory
        rsList = gp.ListRasters("land*", "GRID")
        rs = rsList.Next()
        while rs:
            gp.delete(rs)
            print(rs + " deleted")
            rs = rsList.Next()
        del rs,rsList
        gp.RefreshCatalog(gp.Workspace)

    ####MonoLoop END
    gp.CheckInExtension("Spatial")
    endT = time.time()
    print("Total calculation time is " + str(long(endT-startT)) + " seconds")
    print("rstCA finished now")

except:
    # Print error message if an error occurs
    gp.GetMessages()
